/*
 * Copyright (c) 1997, 2011 LAAS/CNRS -- RIA --
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Based on tip.c written by Matthieu Herrb - Tue Jan  3 1995
 *
 * Maher Khatib - Mon Jun  9 1997
 *
 */

#include <stdio.h>
#include "portLib.h"
#include <sys/time.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <sysLib.h>
#include <errnoLib.h>
#include <taskLib.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <commonStructLib.h>


/*
 * Inclusion
 */
#include "serial.h"

static void* readTask ( void *arg );
static void* writeTask ( void *arg );
static SERIAL_BUFFER_STATE stock ( SERIAL_BUFFER_STR *s, char *data, int size );

#ifndef MIN
#define MIN(x,y) (((x) < (y)) ? (x) : (y))
#endif


/*
 * Read Task
 */

static void *
readTask (void *arg)
{
  fd_set inSet;
  int nFds, n, nLus, aLire;
  char buf[256];
  SERIAL_BUFFER_STR *s = (SERIAL_BUFFER_STR *)arg;
  struct timeval t;

  t.tv_usec = s->period * 1000 / (double)sysClkRateGet();
  t.tv_sec = 0;

  while (1) {
    FD_ZERO(&inSet);
    FD_SET(s->fd, &inSet);
    nFds = select(FD_SETSIZE, &inSet, NULL, NULL, NULL /* &t */);
    if (nFds < 0) {
      perror("serialReadTask: select");
    }
    if (FD_ISSET(s->fd, &inSet)) {
      if (ioctl(s->fd, FIONREAD, &aLire) < 0) {
	perror("serialReadTask: ioctl(FIONREAD)");
      }
#ifdef DEBUG
      fprintf(stderr, "serialReadTask: a lire %d\n", aLire);
#endif
      nLus = 0;
      do {
	n = MIN(aLire, sizeof(buf));
	if ((nLus = read(s->fd, buf, n)) < n) {
	  perror("serialReadTask: read");
	}
	aLire -= nLus;
#ifdef DEBUG
	fprintf(stderr, "serialReadTask: write(%d)\n", nLus);
#endif
	switch (stock (s, buf, nLus)) {
	case SERIAL_BUFFER_ERROR:
	  taskSuspend (0);
	  break;
	case SERIAL_BUFFER_FULL:
	  logMsg("serialReadTask: buffer is full!!\n");
	  /* for debug */
	  serial_flush_in_buffer (s);
	  aLire = 0;
	  break;
	default:
	  break;
	} /* switch */
#ifdef DEBUG
	serial_in_buf_show(s);
#endif
      } while (aLire > 0);
    }
    taskDelay(1);
  } /* while */
  return NULL;
}


/*
 * writeTask
 */

static void *
writeTask (void *arg)
{
  int n, nw;
  SERIAL_BUFFER_STR *s = (SERIAL_BUFFER_STR *)arg;

  while (1) {
    /* Take semaphore */
    if (commonStructTake (s->out.buf) != OK) {
      fprintf(stderr, "serialWriteTask: Cannot take semaphore!\n");
      taskSuspend (0);
    }
    /* Is there any thing to send */
    if (s->out.available != s->out.size) {

      /* compute number of chars */
      n = s->out.size - s->out.available;

      /* check it */
      if (n < 0) {
	fprintf(stderr, "writeTask: Characters number error %d.\n"
	    "\twriteTask is suspended!\n", n);
	s->out.buf_state = SERIAL_BUFFER_ERROR;
	taskSuspend (0);
      }
      /* write them */
      nw = write(s->fd, s->out.buf, n);
#ifdef DEBUG
      fprintf(stderr, "writeTask: tbw %d, wrote %d\n", n, nw);
#endif
      /* scroll buffer */
      if (nw == n) {
	/* update output buffer controls */
	s->out.available = s->out.size;
      } else {
	logMsg ("Buffer Scroll\n");
	/* scroll */
	memcpy (s->out.buf, s->out.buf + nw, n - nw);
	/* update output buffer controls */
	s->out.available += nw;
      }
    } /* if  */

    /* Give semaphore */
    if (commonStructGive (s->out.buf) != OK) {
      fprintf(stderr, "serialWriteTask: Cannot give semaphore!\n");
      /* set error */
      errnoSet (S_serial_cannot_access_semaphore);
      s->out.buf_state = SERIAL_BUFFER_ERROR;
      taskSuspend (0);
    }

    /* wait */
    taskDelay (1);
  } /* while */
  return NULL;
} /* write task */


/*
 * Stock function
 */
static SERIAL_BUFFER_STATE
stock (SERIAL_BUFFER_STR *s, char *data, int size)
{
  /* check state buffer */
  if (s->in.buf_state == SERIAL_BUFFER_FULL) {
    return SERIAL_BUFFER_FULL;
  }

  /* Take semaphore */
  if (commonStructTake (s->in.buf) != OK) {
    return s->in.buf_state = SERIAL_BUFFER_ERROR;
  }

  /* check available size */
  if (size > s->in.available) {
    size = s->in.available;
    s->in.available = 0;
    s->in.buf_state = SERIAL_BUFFER_FULL;
  } else {
    s->in.available -= size;
  }

  /* copy data */
  memcpy (s->in.sp, data, size);

  s->in.sp += size;

  /* Give semaphore */
  if (commonStructGive (s->in.buf) != OK) {
    return s->in.buf_state = SERIAL_BUFFER_ERROR;
  }

  return SERIAL_BUFFER_OK;
} /* stock */


/*
 * Show inout buffer state
 */
int
serial_in_buf_show (SERIAL_BUFFER_STR *s)
{
  char *c;

  /* Take semaphore */
  if (commonStructTake (s->in.buf) != OK) {
    return ERROR;
  }
  printf("Size = %zu\n", s->in.size);
  printf("Available = %zu\n", s->in.available);
  printf("Buf = %p\n", s->in.buf);
  printf("Sp = %p\n", s->in.sp);

  printf("state = %d\n", s->in.buf_state);
  printf("buf:\n");

  c = s->in.buf;

  while (c != s->in.sp) {
    putchar(*c);
    c++;
  } /* while */
  puts("");

  /* Give semaphore */
  if (commonStructGive (s->in.buf) != OK) {
    return ERROR;
  }
  return OK;
} /* serial_in_buf_show */


/*
 * Get number of received chars
 */
int
serial_in_get_n_chars (SERIAL_BUFFER_STR *s)
{
  int n;

  /* Take semaphore */
  if (commonStructTake (s->in.buf) != OK) {
    return ERROR;
  }

  n = s->in.size - s->in.available;

  /* Give semaphore */
  if (commonStructGive (s->in.buf) != OK) {
    return ERROR;
  }

  return n;
}

/*
 * Show output buffer state
 */
int
serial_out_buf_show (SERIAL_BUFFER_STR *s)
{
  char *c;

  /* Take semaphore */
  if (commonStructTake (s->out.buf) != OK) {
    return ERROR;
  }
  printf("Size = %zu\n", s->out.size);
  printf("Available = %zu\n", s->out.available);
  printf("Buf = %p\n", s->out.buf);
  printf("Sp = %p\n", s->out.sp);

  printf("state = %d\n", s->out.buf_state);
  printf("buf:\n");

  c = s->out.buf;

  while (c != s->out.sp) {
    putchar(*c);
    c++;
  } /* while */
  puts("");

  /* Give semaphore */
  if (commonStructGive (s->out.buf) != OK) {
    return ERROR;
  }
  return OK;
}

/*
 * Serial read buffer
 */

int
serial_read_buffer (SERIAL_BUFFER_STR *s, char *buf, int n)
{
  int n_read, n_read_limit;

  /* check n */
  if (n < 1)
    return 0;

  /* check state */
  if (s->in.buf_state == SERIAL_BUFFER_ERROR)
    return ERROR;

  /* Take semaphore */
  if (commonStructTake (s->in.buf) != OK)
    return ERROR;

  /* compute char number */
  n_read = s->in.size - s->in.available;

  /* check */
  if (n_read < 0) {
    /* Give semaphore */
    commonStructGive (s->in.buf);
    return ERROR;
  }

  /* return if nothing to be read */
  if (n_read == 0) {
    /* Give semaphore */
    if (commonStructGive (s->in.buf) != OK)
      return ERROR;
    return 0;
  }

  /* Limit */
  n_read_limit = MIN (n_read, n);

  /* copy data */
  memcpy (buf, s->in.buf, n_read_limit);

  if (n_read_limit < n_read) {
    /* scroll buffer */
    memmove (s->in.buf, s->in.buf + n_read_limit, n_read - n_read_limit);
    s->in.sp -= n_read_limit;
    s->in.available += n_read_limit;
  } else {
    /* update buffer controls */
    s->in.available = s->in.size;
    s->in.sp = s->in.buf;
  }

  /* update state */
  s->in.buf_state = SERIAL_BUFFER_OK;

  /* Give semaphore */
  if (commonStructGive (s->in.buf) != OK)
    return ERROR;

  return n_read_limit;
}

/*
 * Serial scroll buffer
 */
int
serial_scroll_in_buffer (SERIAL_BUFFER_STR *s, int n)
{

  int n_read;

  /* check n */
  if (n < 1)
    return 0;

  /* check buffer state */
  if (s->in.buf_state == SERIAL_BUFFER_ERROR)
    return ERROR;

  /* Take semaphore */
  if (commonStructTake (s->in.buf) != OK)
    return ERROR;

  /* compute char number */
  n_read = s->in.size - s->in.available;

  /* check */
  if (n_read < 0) {
    /* Give semaphore */
    commonStructGive (s->in.buf);
    return ERROR;
  }

  /* check n */
  if (n > n_read) {
    errnoSet (S_serial_scroll_too_big);
    /* Give semaphore */
    commonStructGive (s->in.buf);
    return ERROR;
  }

  /* scroll buffer */
  if (n < n_read) {
    /* scroll buffer */
    memmove (s->in.buf, s->in.buf + n, n_read - n);
    s->in.sp -= n;
    s->in.available += n;
  } else {
    /* update buffer controls */
    s->in.available = s->in.size;
    s->in.sp = s->in.buf;
  }

  /* update state */
  s->in.buf_state = SERIAL_BUFFER_OK;

  /* Give semaphore */
  if (commonStructGive (s->in.buf) != OK)
    return ERROR;

  return n;
}

/*
 * Serial read buffer without scroll
 */

int
serial_read_buffer_without_scroll (SERIAL_BUFFER_STR *s, char *buf, int n)
{
  int n_read, n_read_limit;

  if (s->in.buf_state == SERIAL_BUFFER_ERROR)
    return ERROR;

  /* Take semaphore */
  if (commonStructTake (s->in.buf) != OK)
    return ERROR;

  /* compute char number */
  n_read = s->in.size - s->in.available;

  /* check */
  if (n_read < 0) {
    /* Give semaphore */
    commonStructGive (s->in.buf);
    return ERROR;
  }

  /* return if nothing to be read */
  if (n_read == 0) {
    /* Give semaphore */
    if (commonStructGive (s->in.buf) != OK)
      return ERROR;
    return 0;
  }

  /* Limit */
  n_read_limit = MIN (n_read, n);

  /* copy data */
  memcpy (buf, s->in.buf, n_read_limit);

  /* update state */
  s->in.buf_state = SERIAL_BUFFER_OK;

  /* Give semaphore */
  if (commonStructGive (s->in.buf) != OK)
    return ERROR;

  return n_read_limit;
}

/*
 * Flush input buffer
 */
STATUS
serial_flush_in_buffer(SERIAL_BUFFER_STR *s)
{
  /* Take semaphore */
  if (commonStructTake (s->in.buf) != OK)
    return ERROR;

  s->in.sp = s->in.buf;
  s->in.available = s->in.size;

  /* Flush physic buffer */
  if (tcflush(s->fd, TCIFLUSH) < 0) {
      perror("tcflush");
  }

  /* For debug */
  memset (s->in.buf, 0, s->in.size);

  /* update state */
  s->in.buf_state = SERIAL_BUFFER_OK;

  /* Give semaphore */
  if (commonStructGive (s->in.buf) != OK)
    return ERROR;

  return OK;
}

/*
 * Get input buffer state
 */
SERIAL_BUFFER_STATE
serial_get_in_buffer_state (SERIAL_BUFFER_STR *s)
{
  return s->in.buf_state;
}

/*
 * Get output buffer state
 */
SERIAL_BUFFER_STATE
serial_get_out_buffer_state (SERIAL_BUFFER_STR *s)
{
  return s->out.buf_state;
}

/*
 * Serial write to buffer
 */

int
serial_write_to_buffer (SERIAL_BUFFER_STR *s, char *data, int size)
{

  /* check size */
  if (size > s->out.available)
    size = s->out.available;

  /* Take semaphore */
  if (commonStructTake (s->out.buf) != OK)
    return ERROR;

  /* copy data */
  memcpy (s->out.buf, data, size);
  s->out.available -= size;

  /* Give semaphore */
  if (commonStructGive (s->out.buf) != OK)
    return ERROR;

  /* return size of written chars */
  return size;
}

static struct speed {
    int speed;
    int termioSpeed;
} speeds[] = {
    { 0, B0 },
    { 50, B50 },
    { 75, B75 },
    { 110, B110 },
    { 300, B300 },
    { 600, B600 },
    { 1200, B1200 },
    { 2400, B2400 },
    { 4800, B4800 },
    { 9600, B9600 },
    { 19200, B19200 },
    { 38400, B38400 },
    { 57600, B57600 }};

#define SPEED_SIZE (sizeof(speeds)/sizeof(struct speed))

static int
get_termios_speed(int speed)
{
    int i;

    for (i = 0; i < SPEED_SIZE; i++) {
	if (speed == speeds[i].speed) {
	  return speeds[i].termioSpeed;
	}
    } /* for */
    return -1;
}

/*
 * Change baud rate
 */
STATUS
serial_change_baud_rate (SERIAL_BUFFER_STR *s, int baud)
{
    struct termios tio;

  /* check serial buffer */
  if (!s) {
    errnoSet (S_serial_serial_not_initialized);
    return ERROR;
  }

  /* Take semaphore */
  if (commonStructTake (s->out.buf) != OK)
    return ERROR;

  /* Take semaphore */
  if (commonStructTake (s->in.buf) != OK) {
    return ERROR;
  }

  /* check rate */
  if (baud < SERIAL_BAUD_RATE_MIN || baud > SERIAL_BAUD_RATE_MAX) {
    errnoSet (S_serial_bad_baud_rate);
    return ERROR;
  }

  if (tcgetattr(s->fd, &tio) < 0) {
      perror("tcgetattr");
      return ERROR;
  }
  /* Change baud and flush buffers */
  baud = get_termios_speed(baud);
  if (baud == -1) {
      errnoSet(S_serial_bad_baud_rate);
      return ERROR;
  }
  cfmakeraw(&tio);
  cfsetispeed(&tio, baud);
  cfsetispeed(&tio, baud);

  if (tcsetattr(s->fd, TCSAFLUSH, &tio) < 0) {
      perror("tcsetattr");
      return ERROR;
  }
  /* Give semaphore */
  if (commonStructGive (s->out.buf) != OK)
    return ERROR;

  /* Give semaphore */
  if (commonStructGive (s->in.buf) != OK) {
    return ERROR;
  }

  return OK;
}

/*
 * Init of serial in/out buffers and there control block
 */

SERIAL_BUFFER_STR *
serial_init (char *device, int baud, size_t size_in, size_t size_out,
	     int priority_in, int priority_out)
{
  int serialFd;
  SERIAL_BUFFER_STR *s;
  struct termios tio;

  /* check device */
  if (!device) {
    errnoSet (S_serial_bad_device_name);
    return NULL;
  }

  /* check rate */
  if (baud < SERIAL_BAUD_RATE_MIN || baud > SERIAL_BAUD_RATE_MAX) {
    errnoSet (S_serial_bad_baud_rate);
    return NULL;
  }

  /* open device */
  serialFd = open(device, O_RDWR, 0);

  /* check open */
  if (serialFd == -1) {
    perror(device);
    return NULL;
  }

  baud = get_termios_speed(baud);
  if (baud == -1) {
    errnoSet (S_serial_bad_baud_rate);
    return NULL;
  }

  if (tcgetattr(serialFd, &tio) < 0) {
      perror("tcgetattr");
      return NULL;
  }
  cfmakeraw(&tio);
  cfsetispeed(&tio, baud);
  cfsetospeed(&tio, baud);
  tio.c_cc[VMIN] = 0;
  tio.c_cc[VTIME] = 0;

  if (tcsetattr(serialFd, TCSAFLUSH, &tio) < 0) {
      perror("tcsetattr");
      return NULL;
  }
  /* Allocate control structure */
  if (!(s = malloc (sizeof (SERIAL_BUFFER_STR)))) {
    perror("malloc");
    return NULL;
  }
  /* set fd */
  s->fd = serialFd;

  /* get input buffer size */
  s->in.available = s->in.size = size_in;
  s->in.buf_state = SERIAL_BUFFER_OK;

  /* Alloc protected input buffer */
  if (commonStructCreate (s->in.size, (void **)&s->in.buf) != OK) {
    free (s);
    perror("commonStructCreate");
    return NULL;
  }
  /* Position sp */
  s->in.sp = s->in.buf;

  /* get output buffer size */
  s->out.available = s->out.size = size_out;
  s->out.buf_state = SERIAL_BUFFER_OK;

  /* Alloc protected output buffer */
  if (commonStructCreate (s->out.size, (void **)&s->out.buf) != OK) {
    commonStructDelete (s->in.buf);
    free (s);
    perror("commonStructCreate");
    return NULL;
  }

  s->out.sp = s->out.buf;

  /* Spawn read Task */
  if ((s->in.task_id = taskSpawn2("tSerialRead", priority_in, 0,
	      READ_TASK_STACK_LENGTH, readTask, s))
      == ERROR) {
      perror("taskSpawn");
      commonStructDelete (s->out.buf);
      commonStructDelete (s->in.buf);
      free (s);
      return NULL;
  }

  /* Spawn write Task */
  if ((s->out.task_id = taskSpawn2("tSerialWrite", priority_out, 0,
	      WRITE_TASK_STACK_LENGTH, writeTask, s))
      == ERROR) {
      perror("taskSpawn");
      commonStructDelete (s->out.buf);
      commonStructDelete (s->in.buf);
      free (s);
      return NULL;
  }

  /* return pointer */
  return s;
}

/*
 * End of serial in/out buffers and there control block
 */

STATUS
serial_free (SERIAL_BUFFER_STR *s)
{
  STATUS state = OK;

#ifdef bug
  taskLock();
#endif

  /* free common structs */
  state = commonStructDelete (s->out.buf);
  state = (commonStructDelete (s->in.buf) == OK) ? state : ERROR;

  /* close */
  close (s->fd);

  /* delete Tasks */
  if (taskDelete (s->in.task_id) != OK) {
    fprintf(stderr, "serial_free: Cannot delete readTask!!\n");
    state = ERROR;
  } else {
    printf("serial_free: Delete readTask...OK!!\n");
  }

  if (taskDelete (s->out.task_id) != OK) {
    fprintf(stderr, "serial_free: Cannot delete writeTask!!\n");
    state = ERROR;
  } else {
    printf("serial_free: Delete writeTask...OK!!\n");
  }

  /* free control block */
  free (s);

#ifdef bug
  taskUnlock();
#endif

  return state;
}

/********/
/* Test */
/********/
static int test_stop;
SERIAL_BUFFER_STR *s_test = NULL;

int
serial_test (char *device, int rate)
{
  char buf[256];
  int n;
  int l=0;

  s_test = serial_init (device, rate, 10000, 10000, 5, 5);

  if (!s_test)
    return ERROR;

  printf ("\t\t\tStream = %p\n", s_test);

  test_stop = FALSE;

  while (!test_stop) {
    int i;
    /* read */
    n = serial_read_buffer (s_test, buf, 255);
    l += n;

    for (i = 0; i < n; i++) {
	printf("%02x", buf[i]);
	switch (i % 16) {
	  case 7:
	    printf("  ");
	    break;
	  case 15:
	    printf("\n");
	    break;
	  default:
	    printf(" ");
	}
    }
    if (i%16 != 0) printf("\n");
    taskDelay (1);
  } /* while */

  serial_free (s_test);
  s_test = NULL;

  return OK;
}

void
serial_test_write (char *data)
{
  int n;

  if (!s_test) return;

  if (!data) return;
  n = strlen(data);
  if (n == 0) return;

  if (serial_write_to_buffer (s_test, data, n) != n) {
    fprintf (stderr, "serial_test_write error!!\n");
  }
}

void
serial_test_stop ()
{
  test_stop = TRUE;
}
